import logging
from typing import Tuple, Any, Dict, List
from tablib import Dataset

from im_export.resources import InsureeResource

logger = logging.getLogger(__name__)


class InsureeImportExportService:
    supported_content_types = {
        'xls': 'application/vnd.ms-excel',
        'csv': 'text/csv',
        'json': 'application/json',
        'xlsx': 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet',
    }

    class Strategy:
        INSERT = 'INSERT'
        UPDATE = 'UPDATE'
        INSERT_UPDATE = "INSERT_UPDATE"

    supported_strategies = (Strategy.INSERT, Strategy.UPDATE, Strategy.INSERT_UPDATE)

    def __init__(self, user):
        self._user = user
        self._resource = InsureeResource(user)

    def export_insurees(self, export_format: str = 'csv') -> Tuple[str, Any]:
        if export_format not in self.supported_content_types:
            raise ValueError(f'Non-supported export format: {export_format}')

        # All supported formats match Tablib attrs, to update if that's not valid anymore
        return self.supported_content_types[export_format], \
            getattr(self._resource.export(), export_format)

    def import_insurees(self, import_file, dry_run: bool = False, strategy: str = Strategy.INSERT) \
            -> Tuple[bool, Dict[str, int], List[str]]:

        if not import_file:
            return self._get_general_error('Missing import file')
        if strategy not in self.supported_strategies:
            return self._get_general_error(f'Non-supported strategy: {strategy}')

        # Other strategies are not supported for now
        if strategy in (InsureeImportExportService.Strategy.UPDATE, InsureeImportExportService.Strategy.INSERT_UPDATE):
            strategy = InsureeImportExportService.Strategy.INSERT
            logger.warning(f'Strategy {strategy} not currently supported, defaulting to {InsureeImportExportService.Strategy.INSERT}')

        try:
            if import_file.content_type == 'application/vnd.ms-excel':
                data_set = Dataset(headers=InsureeResource.insuree_headers).load(import_file.open(), 'xls')
            elif import_file.content_type == 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet':
                data_set = Dataset(headers=InsureeResource.insuree_headers).load(import_file.open(), 'xlsx')
            elif import_file.content_type == 'application/json':
                data_set = Dataset(headers=InsureeResource.insuree_headers).load(import_file.open())
            else:
                data_set = Dataset(headers=InsureeResource.insuree_headers).load(import_file.read().decode())
        except Exception as e:
            return self._get_general_error('Failed to parse input file', e)

        try:
            data_set = self._resource.validate_and_sort_dataset(data_set)
        except Exception as e:
            return self._get_general_error('file validation failed', e)

        dry_run_result = self._resource.import_data(data_set, dry_run=True)  # Test the data import
        totals = self._get_totals_from_result(dry_run_result)
        errors = self._get_errors_from_result(dry_run_result)
        success = not dry_run_result.has_errors() and not dry_run_result.has_validation_errors()

        if not dry_run:
            if success:
                self._resource.import_data(data_set, dry_run=False)  # Actually import
                logger.info(f'Imported {totals["sent"]} insurees')
            else:
                logger.info(f'Failed to import {totals["sent"]} insurees, details: {totals}, errors: {errors}')

        return success, totals, errors

    @staticmethod
    def _get_totals_from_result(result):
        return {
            'sent': result.total_rows,
            'created': result.totals['new'],
            'updated': result.totals['update'],
            'deleted': result.totals['delete'],
            'skipped': result.totals['skip'],
            'invalid': result.totals['invalid'],
            'failed': result.totals['error']
        }
        
    @staticmethod
    def _get_general_error(*args):
        errors = []
        for arg in args:
            errors.append(arg.message if hasattr(arg, 'message') else str(arg))
        totals = {'sent': 0, 'created': 0, 'updated': 0, 'deleted': 0, 'skipped': 0, 'invalid': 0, 'failed': 0}
        success = False

        return success, totals, errors

    @staticmethod
    def _get_errors_from_result(result):
        errors = []
        if result.has_validation_errors():
            for invalid_row in result.invalid_rows:
                errors.append(f"row ({invalid_row.number}) - {invalid_row.error.messages}")
        if result.has_errors():
            for index, row_error in result.row_errors():
                for error in row_error:
                    errors.append(f"row ({index}) - {error.error}")
        return errors



from django.apps import AppConfig

MODULE_NAME = "im_export"

# todo these variables should be in settings.py
DEFAULT_CFG = {
    'IMPORT_EXPORT_USE_TRANSACTIONS': False,
    'IMPORT_EXPORT_SKIP_ADMIN_LOG': False,
    'IMPORT_EXPORT_TMP_STORAGE_CLASS': 'import_export.tmp_storages.CacheStorage',
    'IMPORT_EXPORT_IMPORT_PERMISSION_CODE': '131000',
    'IMPORT_EXPORT_EXPORT_PERMISSION_CODE': '131000',
    'IMPORT_EXPORT_CHUNK_SIZE': '100',

    'im_export_date_format': '%m/%d/%Y',
}


class ImportExportConfig(AppConfig):
    name = MODULE_NAME

    # IMPORT_EXPORT_USE_TRANSACTIONS = False
    # IMPORT_EXPORT_SKIP_ADMIN_LOG = False
    # IMPORT_EXPORT_TMP_STORAGE_CLASS = None
    # IMPORT_EXPORT_IMPORT_PERMISSION_CODE = None
    # IMPORT_EXPORT_CHUNK_SIZE = 100

    im_export_date_format = None

    def _configure_permissions(self, cfg):
        # managed by django-import-export
        pass

    def ready(self):
        from core.models import ModuleConfiguration
        cfg = ModuleConfiguration.get_or_default(MODULE_NAME, DEFAULT_CFG)
        self._configure_permissions(cfg)

        self.im_export_date_format = cfg['im_export_date_format']



from django.db.models import Q

from core.datetimes.ad_datetime import datetime
from im_export.apps import ImportExportConfig
from insuree.models import Insuree, Family, Gender
from location.models import Location

from import_export import fields, resources, widgets


class CharRequiredWidget(widgets.CharWidget):
    def clean(self, value, row=None, *args, **kwargs):
        val = super().clean(value)
        if val:
            return val
        else:
            raise ValueError('this field is required')


class ForeignkeyRequiredWidget(widgets.ForeignKeyWidget):
    def clean(self, value, row=None, *args, **kwargs):
        if value:
            return self.get_queryset(value, row, *args, **kwargs).get(**{self.field: value})
        else:
            raise ValueError(self.field + ' required')


def get_location_str_filter(str):
    return Q(code=str) | Q(name=str)  # | Q(uuid=str) | Q(id=int(str) if str.isdigit() else None)


def get_locations_ids(row, regions, districts, municipalities, villages):
    uniq_region = str(row['region'] or '').strip()
    if uniq_region not in regions:
        region_model = Location.objects.all().filter(validity_to__isnull=True) \
            .filter(get_location_str_filter(uniq_region)).first()
        if region_model:
            regions[uniq_region] = region_model.id
        else:
            raise ValueError('Location {} not found in the database'.format(uniq_region))

    district = str(row['district'] or '').strip()
    uniq_district = uniq_region + "|" + district
    if uniq_district not in districts:
        district_model = Location.objects.all() \
            .filter(validity_to__isnull=True, parent__id=regions[uniq_region]) \
            .filter(get_location_str_filter(district)).first()
        if district_model:
            districts[uniq_district] = district_model.id
        else:
            raise ValueError('Location {} not found in the database'.format(uniq_district))

    municipality = str(row['municipality'] or '').strip()
    uniq_municipality = uniq_district + "|" + municipality
    if uniq_municipality not in municipalities:
        municipality_model = Location.objects.all() \
            .filter(validity_to__isnull=True, parent__id=districts[uniq_district]) \
            .filter(get_location_str_filter(municipality)).first()
        if municipality_model:
            municipalities[uniq_municipality] = municipality_model.id
        else:
            raise ValueError('Location {} not found in the database'.format(uniq_municipality))

    village = str(row['village'] or '').strip()
    uniq_village = uniq_municipality + "|" + village
    if uniq_village not in villages:
        village_model = Location.objects.all() \
            .filter(validity_to__isnull=True, parent__id=municipalities[uniq_municipality]) \
            .filter(get_location_str_filter(village)).first()
        if village_model:
            villages[uniq_village] = village_model.id
        else:
            raise ValueError('Location {} not found in the database'.format(uniq_village))

    return regions[uniq_region], districts[uniq_district], municipalities[uniq_municipality], villages[uniq_village]


def validate_and_preprocess(dataset):
    # change the region names/codes into id.
    # a village can have homonyms therefore the parent must be taken into account
    # check if head insuree
    # check if insuree number is unique
    regions = dict()
    districts = dict()
    municipalities = dict()
    villages = dict()

    regions_ids = list()
    districts_ids = list()
    municipalities_ids = list()
    villages_ids = list()
    head = list()
    empty_indices = list()

    insuree_no_seen = set()

    for idx, row in enumerate(dataset.dict):

        row_str = ''.join([str(col or '') for col in row])
        if row_str.strip() == '':
            empty_indices.append(idx)
        else:
            # Check locations
            region_id, district_id, municipality_id, village_id = \
                get_locations_ids(row, regions, districts, municipalities, villages)
            regions_ids.append(region_id)
            districts_ids.append(district_id)
            municipalities_ids.append(municipality_id)
            villages_ids.append(village_id)

            insuree_number = row['insuree_number']
            if insuree_number not in insuree_no_seen:
                insuree_no_seen.add(insuree_number)

            if insuree_number == row['head_insuree_number']:
                head.append(True)
            else:
                head.append(False)

    for index in sorted(empty_indices, reverse=True):
        del dataset[index]

    if len(insuree_no_seen) != len(dataset['insuree_number']):
        raise ValueError('There are duplicates of insurance number in the list')

    dataset.append_col(head, 'head')
    dataset.append_col(villages_ids, 'village_id')
    dataset.append_col(municipalities_ids, 'municipality_id')
    dataset.append_col(districts_ids, 'district_id')
    dataset.append_col(regions_ids, 'region_id')


class InsureeResource(resources.ModelResource):
    insuree_headers = ['head_insuree_number', 'insuree_number', 'last_name', 'other_names', 'dob', 'sex', 'village',
                       'municipality', 'district', 'region']

    head_insuree_number = fields.Field(
        attribute='family',
        column_name='head_insuree_number',
        widget=ForeignkeyRequiredWidget(Family, field='head_insuree__chf_id'),
        readonly=True,
    )

    insuree_number = fields.Field(
        attribute='chf_id',
        column_name='insuree_number',
    )

    last_name = fields.Field(
        attribute='last_name',
        column_name='last_name',
    )

    other_names = fields.Field(
        attribute='other_names',
        column_name='other_names',
    )

    dob = fields.Field(
        attribute='dob',
        column_name='dob',
        widget=widgets.DateWidget(format=ImportExportConfig.im_export_date_format)
    )

    sex = fields.Field(
        attribute='gender',
        column_name='sex',
        widget=ForeignkeyRequiredWidget(Gender, field='code'),
    )

    village = fields.Field(
        attribute='current_village',
        column_name='village',
        widget=ForeignkeyRequiredWidget(Location, field='name'),
        readonly=True
    )

    # readonly, just for export and import validation
    municipality = fields.Field(
        attribute='current_village',
        column_name='municipality',
        widget=ForeignkeyRequiredWidget(Location, field='parent__name'),
        readonly=True
    )

    # readonly, just for export and import validation
    district = fields.Field(
        attribute='current_village',
        column_name='district',
        widget=ForeignkeyRequiredWidget(Location, field='parent__parent__name'),
        readonly=True
    )

    # readonly, just for export and import validation
    region = fields.Field(
        attribute='current_village',
        column_name='region',
        widget=ForeignkeyRequiredWidget(Location, field='parent__parent__parent__name'),
        readonly=True
    )

    def __init__(self, user, queryset=None, ):
        """
        @param user: User to be used for location rights for import and export, and for audit_user_id
        @param queryset: Queryset to use for export, Default to full quetyset
        """
        super().__init__()
        self._user = user
        self._queryset = queryset

    @classmethod
    def validate_and_sort_dataset(cls, dataset):
        validate_and_preprocess(dataset)
        return dataset.sort('head', reverse=True)

    def after_import_instance(self, instance, new, **kwargs):
        # for now editing is disabled
        if not new:
            raise ValueError("Insuree number already exists")

    def import_obj(self, instance, row, dry_run, **kwargs):
        instance.head = row['head']
        instance.current_village_id = row['village_id']

        if not instance.id:
            instance.card_issued = False
            instance.audit_user_id = self._user.i_user.id

        if not instance.head:
            family = Family.objects.all().filter(validity_to__isnull=True) \
                .select_related('location') \
                .get(head_insuree__chf_id=row['head_insuree_number'])
            instance.family = family
            family_location = family.location
            # if the current village is the same as head's, skip the current_location
            if family_location.id == row['village_id']:
                instance.current_village_id = None

        # important to be at the end
        super().import_obj(instance, row, dry_run, **kwargs)

    def after_save_instance(self, instance, using_transactions, dry_run):
        super().after_save_instance(instance, using_transactions, dry_run)
        if instance.head:
            # if not using_transactions and dry_run this code will cause changes in database on dry run
            if using_transactions or not dry_run:
                instance.family = self.create_family(instance)
                instance.current_village = None
                instance.save()

    def create_family(self, instance):
        return Family.objects.create(**{
            'validity_from': datetime.now(),
            'audit_user_id': self._user.i_user.id,
            'head_insuree': instance,
            'location': instance.current_village,
            'is_offline': False,
        })

    def get_queryset(self):
        queryset = self._queryset if self._queryset else super().get_queryset()
        return queryset \
            .filter(validity_to__isnull=True) \
            .select_related('gender', 'current_village', 'family', 'family__location', 'family__location__parent',
                            'family__location__parent__parent', 'family__location__parent__parent__parent')

    class Meta:
        model = Insuree
        import_id_fields = ('insuree_number',)
        fields = ('insuree_number', 'head_insuree_number', 'last_name', 'other_names', 'dob', 'sex', 'village',
                  'municipality', 'district', 'region')



import json
import logging
from django.http import HttpResponse, JsonResponse
from rest_framework.decorators import api_view, permission_classes
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response
from insuree.apps import InsureeConfig
from .services import InsureeImportExportService

logger = logging.getLogger(__name__)


def check_user_rights(rights):
    class UserWithRights(IsAuthenticated):
        def has_permission(self, request, view):
            return super().has_permission(request, view) and request.user.has_perms(rights)

    return UserWithRights


@api_view(["POST"])
@permission_classes([check_user_rights(InsureeConfig.gql_mutation_create_insurees_perms, )])
def import_insurees(request):
    try:
        import_file = request.FILES.get('file', None)
        user = request.user
        dry_run = json.loads(request.POST.get('dry_run', 'false'))
        strategy = request.POST.get('strategy', InsureeImportExportService.Strategy.INSERT)

        success, totals, errors = InsureeImportExportService(user) \
            .import_insurees(import_file, dry_run=dry_run, strategy=strategy)
        return JsonResponse(data={'success': success, 'data': totals, 'errors': errors})
    except ValueError as e:
        logger.error("Error while importing insurees", exc_info=e)
        return Response({'success': False, 'error': str(e)}, status=400)
    except Exception as e:
        logger.error("Unexpected error while importing insurees", exc_info=e)
        return Response({'success': False, 'error': str(e)}, status=500)


@api_view(["GET"])
@permission_classes([check_user_rights(InsureeConfig.gql_query_insurees_perms, )])
def export_insurees(request):
    try:
        # TODO add location based filtering
        export_format = request.GET.get("file_format", "csv")
        user = request.user

        content_type, export = InsureeImportExportService(user).export_insurees(export_format)
        response = HttpResponse(export, content_type=content_type)
        response['Content-Disposition'] = f'attachment; filename="insurees.{export_format}"'
        return response
    except ValueError as e:
        logger.error("Error while exporting insurees", exc_info=e)
        return Response({'success': False, 'error': str(e)}, status=400)
    except Exception as e:
        logger.error("Unexpected error while exporting insurees", exc_info=e)
        return Response({'success': False, 'error': str(e)}, status=500)



import os

from setuptools import find_packages, setup

with open(os.path.join(os.path.dirname(__file__), 'README.md')) as readme:
    README = readme.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='openimis-be-im_export',
    version='v1.2.0',
    packages=find_packages(),
    include_package_data=True,
    license='GNU AGPL v3',
    description='The openIMIS Backend import-export reference module based on django-import-export.',
    # long_description=README,
    url='https://openimis.org/',
    author='Patrick Delcroix',
    author_email='patrick.delcroix@swisstph.ch',
    install_requires=[
        'django',
        'django-import-export',
        'tablib',
    ],
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 3.0',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Affero General Public License v3',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.9',
    ],
)



import os
from tablib import Dataset
from im_export.resources import InsureeResource
from core.services import create_or_update_core_user, create_or_update_interactive_user
from django.test import TestCase
from location.test_helpers import create_test_location

_TEST_USER_NAME = "test_insuree_import"
_TEST_USER_PWD = "test_insuree_import"
_TEST_DATA_USER = {
    "username": _TEST_USER_NAME,
    "last_name": _TEST_USER_NAME,
    "password": _TEST_USER_PWD,
    "other_names": _TEST_USER_NAME,
    "user_types": "INTERACTIVE",
    "language": "en",
    "roles": [1, 5, 9],
}

# all location used in the test files must use those name
_TEST_LOCATIONS = [
    {
        'region': 'Batha',
        'district': 'Batha',
        'municipality': 'BEGOU',
        'villages': [
            'Baguirmi',
            'Boua',
            'Niellim',
            'Sarakaba',
            'Maroc'
        ]
    }
]


class ImportInsureeTest(TestCase):

    def setUp(self) -> None:

        super(ImportInsureeTest, self).setUp()
        self.i_user, i_user_created = create_or_update_interactive_user(
            user_id=None, data=_TEST_DATA_USER, audit_user_id=999, connected=False)
        self.user, user_created = create_or_update_core_user(
            user_uuid=None, username=_TEST_DATA_USER["username"], i_user=self.i_user)
        test_location = _TEST_LOCATIONS

        for locations in test_location:
            test_region = create_test_location('R', custom_props={"name": locations['region']})
            test_district = create_test_location('D', custom_props={"name": locations['district'],
                                                                    "parent_id": test_region.id})
            test_municipality = create_test_location('M', custom_props={"name": locations['municipality'],
                                                                        "parent_id": test_district.id})
            for villages in locations['villages']:
                create_test_location('V', custom_props={"name": villages, "parent_id": test_municipality.id})

    def test_simple_import(self):
        dir_path = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
        resource = InsureeResource(user=self.user)
        with open(os.path.join(dir_path, 'tests/import_example.csv'), 'r') as f:
            imported_data = resource \
                .validate_and_sort_dataset(Dataset(headers=InsureeResource.insuree_headers).load(f.read()))
            result = resource.import_data(
                imported_data, dry_run=True, use_transactions=True,
                collect_failed_rows=False,
            )
            self.assertEqual(result.has_errors(), False)

    def test_simple_export(self):
        result = InsureeResource(self.user).export().dict
        self.assertTrue(result)

# todo expand tests
